// DISB-USB project 1. : This file contains the 'main' function. 
//


/* This cpp file is meant to provide a demonstration of how the DISB-USB bridge in conjuction
with an Ibsen Photonics spectrometer, can be operated.
It is up to the individual use to adapt the script to their needs.

Running this script will estblish a connection to the DISB hardware, perform read/write to the majority of the
registers, read calibration coefficients stored on the hardware and save it to a text file (example.txt).
Two spectra will the be taken, one with 10 ms integration time and one with 100 ms intration time.

More information consult the Application Note AN_329 - User Guide for LibFT4222 Manual, which can be downloaded from FTDI's website.
https://www.ftdichip.com/Products/ICs/FT4222H.html

*/


/*Inclusion of header, .ddl and liberary files should all be included in the project.
The files can be downloaded from:
https://www.ftdichip.com/Products/ICs/FT4222H.html
*/
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <array>
#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <stdexcept>
#include <ctime>
#include "LibFT4222.h"
#include "ftd2xx.h"

using namespace std;

// Setting up strings to be returned 
inline std::string DeviceFlagToString(DWORD flags)
{
	std::string msg;
	msg += (flags & 0x1) ? "DEVICE_OPEN" : "DEVICE_CLOSED";
	msg += ", ";
	msg += (flags & 0x2) ? "High-speed USB" : "Full-speed USB";
	return msg;
}

//Function to list number of USB devices. 
//There are 4 different USB serial number of A, B, C and D, these are not to be confused with the revision types of the chip also denoted A, B, C and D. 
//The FT4222H chip can be set to either of 4 Chip Mode Configuaration (0-3). (Read more in FT4222H Datasheet, https://www.ftdichip.com/Products/ICs/FT4222H.html)
//The DISB-USB Bridge is hardwired with both the DCNF0 and DCNF1 pins, set high, and the FT4222H is subsequently in MODE1.
/*
CNFMODE0 = (DCNF1, DCNF0) 00
CNFMODE1 = (DCNF1, DCNF0) 01
CNFMODE2 = (DCNF1, DCNF0) 10
CNFMODE3 = (DCNF1, DCNF0) 11


Pin Functions		| CNFMODE0 = (DCNF1, DCNF0) 00	| CNFMODE1 = (DCNF1, DCNF0) 01	| CNFMODE2 = (DCNF1, DCNF0) 10	| CNFMODE3 = (DCNF1, DCNF0) 11	|
USB inferface number| 1 for Datastream 1 for GPIOs	| 3 for Datastreams 1 for GPIOs	|  4 for Datastreams			| 1 for Datastream				|
SPI Master (SPIM)	|				SPIM*			|	Active USB interface 0		|	Active USB interface 0		|				SPIM*			|
SPI Slave (SPIS)	|				SPIS*			|			Disable				|			Disable				|				SPIS*			|
GPIO0				|			GPIO/SCL*			|	SS1O USB interface-1		|	SS1O USB interface-1		|				SCL*			|
GPIO1				|			GPIO/SDA*			|	SS2O USB interface-2		|	SS2O USB interface-2		|				SDA*			|
GPIO2				|			GPIO/SUSP*			|			GPIO/SUSP			|	SS3O USB interface-3		|				SUSP			|
GPIO3				|			GPIO/WAKE/INTR		|			GPIO/WAKE/INTR		|			WAKE				|				WAKE			|



*/


std::vector< FT_DEVICE_LIST_INFO_NODE > g_FT4222DevList;

void ListFtUsbDevices()
{
	//Create status and number of devices.
	FT_STATUS ftStatus = 0;
	DWORD numOfDevices = 0;
	printf("\n%x " ,&numOfDevices);
	printf("\n%x ", numOfDevices);
	//Create a List for the Device info to be sent to. 
	//There should be 4 device, Serial A, B, C and D.
	ftStatus = FT_CreateDeviceInfoList(&numOfDevices);

	//Cycle through the different devices.
	for (DWORD iDev = 0; iDev < numOfDevices; ++iDev)
	{
		FT_DEVICE_LIST_INFO_NODE devInfo;
		memset(&devInfo, 0, sizeof(devInfo));

		//Get information about device number "iDev"
		ftStatus = FT_GetDeviceInfoDetail(iDev, &devInfo.Flags, &devInfo.Type, &devInfo.ID, &devInfo.LocId,
			devInfo.SerialNumber,
			devInfo.Description,
			&devInfo.ftHandle);

		//If a device was opened correctly, Hence ftStatus == FT_OK, Print the device information.
		if (FT_OK == ftStatus)
		{
			printf("Dev %d:\n", iDev);
			printf("  Flags= 0x%x, (%s)\n", devInfo.Flags, DeviceFlagToString(devInfo.Flags).c_str());
			printf("  Type= 0x%x\n", devInfo.Type);
			printf("  ID= 0x%x\n", devInfo.ID);
			printf("  LocId= 0x%x\n", devInfo.LocId);
			printf("  SerialNumber= %s\n", devInfo.SerialNumber);
			printf("  Description= %s\n", devInfo.Description);
			printf("  ftHandle= 0x%p\n", devInfo.ftHandle);

			const std::string desc = devInfo.Description;
			if (desc == "FT4222" || desc == "FT4222 A")
			{
				g_FT4222DevList.push_back(devInfo);
			}
		}
	}
	printf("\n%x ", &numOfDevices);
	printf("\n%x ", numOfDevices);
}

uint16_t ReadRegister(PVOID FThandle, int RegisterAddress)
{
	uint8_t AppendedRegAddres = RegisterAddress * 4 + 2;

	//Create a read- and writebuffer for data to be read and writen to the DISB registers.
	uint8 ReadBuffer[3], WriteBuffer[3];
	//Used as the adress for which the number of bytes being read or writen after each ReadWrite command is registed.
	uint16 SizeTransferred;

	WriteBuffer[0] = AppendedRegAddres;
	WriteBuffer[1] = 0x0;
	WriteBuffer[2] = 0x0;

	FT4222_SPIMaster_SingleReadWrite(FThandle, ReadBuffer, WriteBuffer, 3, &SizeTransferred, 1);
	uint16_t RegisterValue = (ReadBuffer[1] << 8) | (ReadBuffer[2] & 0xff);
	return RegisterValue;
}
int SetRegister(PVOID FThandle, int RegisterAddress, int NewRegisterValue)
{
	uint8_t AppendedRegAddres = RegisterAddress * 4;
	
	//Create a read- and writebuffer for data to be read and writen to the DISB registers.
	uint8 ReadBuffer[3], WriteBuffer[3], MSB, LSB;
	//Used as the adress for which the number of bytes being read or writen after each ReadWrite command is registed.
	uint16 SizeTransferred;

	WriteBuffer[0] = AppendedRegAddres;
	MSB = NewRegisterValue >> 8;
	LSB = NewRegisterValue;

	WriteBuffer[1] = MSB;
	WriteBuffer[2] = LSB;


	FT4222_SPIMaster_SingleReadWrite(FThandle, ReadBuffer, WriteBuffer, 3, &SizeTransferred, 1);

	return 0;
}

int main(int argc, char const* argv[])
{

	/*	*********************************************************************
		*********************************************************************
		Search for available Device adapters
		*********************************************************************
		*********************************************************************
	*/
	ListFtUsbDevices();

	//If no device could be found print a "No FT4222 device is found!" message to terminal.
	if (g_FT4222DevList.empty()) {
		printf("No FT4222 device is found!\n");
		return 0;
	}

	printf("\n-----------------------------------------------------------------------------------------------");
	printf("\n-----------------------------------------------------------------------------------------------");
	/*	*********************************************************************
		*********************************************************************
		Opening Device FT4222
		*********************************************************************
		*********************************************************************
	*/
	//Set up a device Handle and Status for device.
	FT_HANDLE ftHandleOutStream = NULL;
	FT_STATUS ftStatus;
	

	//Open device with serial number A. 
	//ftHandleOutStream is used for bulk data out stream of image
	ftStatus = FT_OpenEx("A", FT_OPEN_BY_SERIAL_NUMBER, &ftHandleOutStream);
	//If the device was not opened correctly return an error message.
	if (FT_OK != ftStatus) {
		printf("Open a FT4222 device failed!\n");
		return 0;
	}
	
	//Handle for the CS0 (DISB HW Manual) dataline.
	FT_HANDLE ftHandleOutCom = NULL;

	//Open device by serial number B
	//ftHandleOutCom is used for read write data, with the DISB registers, serial numbers, integration time, trigger delays and so on. 
	ftStatus = FT_OpenEx("B", FT_OPEN_BY_SERIAL_NUMBER, &ftHandleOutCom);
	if (FT_OK != ftStatus) {
		printf("\nOpen a FT4222 device failed!\n");
		return 0;
	}

	//Handle for the GPIO.
	FT_HANDLE ftHandleOutGPIO = NULL;

	//Open device by serial number C, to be used for GPIOs 
	ftStatus = FT_OpenEx("C", FT_OPEN_BY_SERIAL_NUMBER, &ftHandleOutGPIO);
	if (FT_OK != ftStatus) {
		printf("Open a FT4222 device failed!\n");
		return 0;
	}



	/*	
	*********************************************************************
	*********************************************************************
	Initialize Device Master SPI
	*********************************************************************
	*********************************************************************
	*/
	//-------------------------------------------------------------------------------------------------------------
	// SET LATENCY TIMER
	//-------------------------------------------------------------------------------------------------------------
	UCHAR UcLatency = 20;
	ftStatus = FT_SetLatencyTimer(ftHandleOutStream, UcLatency);

	//-------------------------------------------------------------------------------------------------------------
	// SET USB PARAMETERS 
	//-------------------------------------------------------------------------------------------------------------
	ULONG OutTransferSize = 4096;
	ULONG InTransferSize = 65536;
	ftStatus = FT_SetUSBParameters(ftHandleOutStream, InTransferSize, OutTransferSize);

	//-------------------------------------------------------------------------------------------------------------
	// SET CLOCK  - 60 MHz
	//-------------------------------------------------------------------------------------------------------------

	FT4222_SetClock(ftHandleOutStream, SYS_CLK_60);

	//-------------------------------------------------------------------------------------------------------------
	// MASTER SPI INTIALIZE 

	//SPI Status
	FT4222_STATUS ft4222Status;

	//Initilize the FT4222H as a master for USB to SPI bridge functionality. 
	/*
	Handle,
	One bit at a time data stream
	Divide clock by 4 resulting in 15MHz SPI clock.
	Polarity = 0
	Phase = 1
	Slave Select SS (Triggered low) SS1O
	*/
	ft4222Status = FT4222_SPIMaster_Init(ftHandleOutCom, SPI_IO_SINGLE, CLK_DIV_4, CLK_IDLE_LOW, CLK_TRAILING, 2);
	if (FT4222_OK != ft4222Status) {
		printf("Init FT4222 as SPI master device failed!\n");
		return 0;
	}
	else {
		//Print the pointer of the Read write register handle, if open correctly.
		printf("\nPointer of Bulk data handle, CS0 in DISB HW manual : % p", ftHandleOutCom);
	}


	//Initilize the FT4222H as a master for USB to SPI bridge functionality. 
	/*
	Handle,
	One bit at a time data stream
	Divide clock by 4 resulting in 15MHz SPI clock.
	Polarity = 0
	Phase = 1
	Slave Select SS (Triggered low) SS0O  
	*/
	ft4222Status = FT4222_SPIMaster_Init(ftHandleOutStream, SPI_IO_SINGLE, CLK_DIV_4, CLK_IDLE_LOW, CLK_TRAILING, 3);
	if (FT4222_OK != ft4222Status) {
		printf("Init FT4222 as SPI master device failed!\n");
		return 0;
	}
	else {
		//Print the pointer of the Bulk data handle, if open correctly.
		printf("\nPointer of Register parameter handle, CS1 in DISB HW manual: %p", ftHandleOutStream);
	}


	//Setting up GPIO 
	FT_SetLatencyTimer(ftHandleOutGPIO, UcLatency);
	GPIO_Dir gpioDir[4];
	FT4222_GPIO_Init(ftHandleOutGPIO, gpioDir);



	/*	
	*********************************************************************
	*********************************************************************
	Read parameter from DISB Register
	*********************************************************************
	*********************************************************************
	*/
	//Create a text --  example.txt and open it.
	ofstream Textfile;
	Textfile.open("example.txt");

	//Perform a single read of register 1, and both print the value to the terminal and example.txt
	uint16_t PCB_SN = ReadRegister(ftHandleOutCom, 1);
	printf("\nPCB SERIAL NUMBER %d", PCB_SN);
	Textfile << "PCB_SN: ";
	Textfile << PCB_SN;



	//Read the HW version of the DISB board, located in register 2. 
	uint16_t HardwareVersion = ReadRegister(ftHandleOutCom, 2);
	printf("\nHW_TYPE %d", HardwareVersion);
	Textfile << "\nHardware version: ";
	Textfile << HardwareVersion;



	//Read the FW version of the DISB board, located in register 3. 
	uint16_t FirmwareVersion = ReadRegister(ftHandleOutCom, 3);
	printf("\nFIRMWARE %d", FirmwareVersion);
	Textfile << "\nFirmware version: ";
	Textfile << FirmwareVersion;



	//Read the Detector type, located in register 4. 
	uint16_t DetectorType = ReadRegister(ftHandleOutCom, 4);
	printf("\nDetector type: %d", DetectorType);
	Textfile << "\nDetector type: ";
	Textfile << DetectorType;



	//Read the number of pixels per image, located in register 5. 
	uint16_t PixelPerImage = ReadRegister(ftHandleOutCom, 5);
	printf("\nPixel per Image: %d", PixelPerImage);
	Textfile << "\nPixel per Image: ";
	Textfile << PixelPerImage;



	//Read the number of characters used for calibration coeffients, located in register 6. 
	uint16_t NumberofCaliChars = ReadRegister(ftHandleOutCom, 6);
	printf("\nNumber of characters of calibration coeffients: %d ", NumberofCaliChars);
	Textfile << "\nNumber of characters of calibration coeffients: ";
	Textfile << NumberofCaliChars;

	

	//Read the wavelength calibration coeffients, via register 7.
	//When register 6 has been read the pointer is directed toward the first address of register 7.
	//Reading register 7 automatically increments the pointer. 
	printf("\nCalibration coeffients: \n");
	Textfile << "\nCalibration coeffients: \n";
	array<CHAR, 14> calibrationChars;
	string CombinedCalibrationChars;
	for (int i = 0; i < NumberofCaliChars / 14; i++) 
	{
		CombinedCalibrationChars = "";
		for (int j = 0; j < 14; j++)
		{
			calibrationChars[j] = ReadRegister(ftHandleOutCom, 7);
			CombinedCalibrationChars = CombinedCalibrationChars + calibrationChars[j];
		}
		cout << CombinedCalibrationChars << endl;
		Textfile << CombinedCalibrationChars << endl;
	}


	//Read the temperature register number 11. 
	uint16_t Temperature = ReadRegister(ftHandleOutCom, 11);
	printf("Temperature measured: %d", Temperature);
	Textfile << "\nTemperature measured: ";
	Textfile << Temperature;


	//Read the Trigger delay, located in register 13 and 14. 
	uint16_t TriggerDelayLSB = ReadRegister(ftHandleOutCom, 13);
	uint16_t TriggerDelayMSB = ReadRegister(ftHandleOutCom, 14);
	uint32_t TriggerDelay = (TriggerDelayLSB << 16) | (TriggerDelayMSB & 0xffff);
	Textfile << "\nTrigger delay: ";
	Textfile << TriggerDelay;



	//Read the Amplifier Gain, located in register 15. 
	uint16_t ADCGain = ReadRegister(ftHandleOutCom, 15);
	printf("\nADC Programmable Gain Amplification (PGA) number: %d", ADCGain);
	Textfile << "\nADC Programmable Gain Amplification (PGA) number:  ";
	Textfile << ADCGain;



	//Read the Amplifier offset, located in register 16. 
	uint16_t ADCOffset = ReadRegister(ftHandleOutCom, 16);
	printf("\nADC offset number: %d", ADCOffset);
	Textfile << "\nADC offset number: ";
	Textfile << ADCOffset;




	//Read the Spectrometer serial number, located in register 22. 
	uint16_t SpectrometerSN = ReadRegister(ftHandleOutCom, 22);
	printf("\nSpectrometer serial number: %d", SpectrometerSN);
	Textfile << "\nSpectrometer serial number: ";
	Textfile << SpectrometerSN;




	//Read the first pixel to be read, located in register 23. 
	uint16_t FirstPixel = ReadRegister(ftHandleOutCom, 23);
	printf("\nFirst pixel read: %d", FirstPixel);
	Textfile << "\nFirst pixel read: ";
	Textfile << FirstPixel;




	//Read the last pixel to be read, located in register 24. 
	uint16_t LastPixel = ReadRegister(ftHandleOutCom, 24);
	printf("\nLast pixel read: %d", LastPixel);
	Textfile << "\nLast pixel read: ";
	Textfile << LastPixel;

	Textfile << "\n----------------------------------------------------------------------------------------------- ";
	Textfile << "\n----------------------------------------------------------------------------------------------- ";

	printf("\n-----------------------------------------------------------------------------------------------");
	printf("\nSpectrum with 10 ms integration time");
	printf("\n-----------------------------------------------------------------------------------------------");
	
	//-------------------------------------------------------------------------------------------------------------------------------
	/*
	Examples of setting different integration times and reading spectra
	*/
	//-------------------------------------------------------------------------------------------------------------------------------

	/*	*********************************************************************
	*********************************************************************
	Write integration time to DISB Register
	*********************************************************************
	*********************************************************************
	*/

	/*
	Setting an integration time of 10 ms.
	Integration time is set in increments of 200 ns, so 10 ms => 50.000 => MSB 0x000 - LBS 0xC350
	*/
	// Setting integration time of Register 10 MSB. 
	SetRegister(ftHandleOutCom, 10, 0);

	// Setting integration time of Register 9 LSB. 
	SetRegister(ftHandleOutCom, 9, 50000);

	// Read the integration time of register 10 MSB. 
	uint16_t MSB_10ms = ReadRegister(ftHandleOutCom, 10);
	// Read the integration time of register 9 MSB. 
	uint16_t LSB_10ms = ReadRegister(ftHandleOutCom, 9);

	uint32_t IntegrationTime = (MSB_10ms << 16) | (LSB_10ms & 0xffff);
	

	Textfile << "\nIntegration time in register in counts: ";
	Textfile << IntegrationTime;
	Textfile << "\nIntegration time in register in milliseconds: ";
	Textfile << IntegrationTime * 0.0002;
	Textfile << " [ms]";
	//-------------------------------------------------------------------------------------------------------------------------------
	
	/*	
	*********************************************************************
	*********************************************************************
	Trigger exposure via write to Sensor CTRL register.
	*********************************************************************
	*********************************************************************
	*/

	// Writing to the Sensor Control register 8.
	// Perform soft reset of the image transfer buffer
	SetRegister(ftHandleOutCom, 8, 0x10);

	//sleep_for(milliseconds(100));
	// Trigger an exposure
	SetRegister(ftHandleOutCom, 8, 0x01);

	/*
	*********************************************************************
	*********************************************************************
	Read spectrum from framebuffer.
	*********************************************************************
	*********************************************************************
	*/
	//-------------------------------------------------------------------------------------------------------------------------------
	// Change the chip select from 0 to 1 (CS0 --> CS1) and Framesize from 8-bit bytes to 16-bit
	Textfile << "\n----------------------------------------------------------------------------------------------- ";
	Textfile << "\n";
	printf("\n");

	//FrameBuffer size is set to double the amount of pixel, given that the data is read in 8-bit bytes, but the ADC is 16-bit.
	//The two bytes are then stiched together to form the correct pixel values.
	uint16 datalength = (LastPixel - FirstPixel + 1.0) * 2.0;
	std::vector<uint8> NewReadFrambuffer{ 0 };
	NewReadFrambuffer.resize(datalength);
	std::vector<uint8> NewWriteFrambuffer{ 0 };
	NewWriteFrambuffer.resize(datalength);
	
	//Wait for the framebuffer to be full (NumberOfPixReady = LastPixel) or for timeout after 1 second.
	clock_t StartTime = clock();
	clock_t CurrentTime = clock();
	int TimeElapsedinSec;
	uint16_t NumberOfPixReady;
	do
	{
		NumberOfPixReady = ReadRegister(ftHandleOutCom, 12);
		CurrentTime = clock();
		TimeElapsedinSec = (CurrentTime - StartTime) / CLOCKS_PER_SEC;
	} while (LastPixel - NumberOfPixReady != 0 && TimeElapsedinSec < 1);
	if (TimeElapsedinSec >= 1)
	{
		printf("\nTimeout error when capturing spectrum\n");
	}

	uint16_t SizeTransferred;
	//Read the pixel values
	FT4222_SPIMaster_SingleReadWrite(ftHandleOutStream, NewReadFrambuffer.data(), NewWriteFrambuffer.data(), datalength, &SizeTransferred, 1);


	//Stich and print the pixel values
	uint16_t SpectromterPix;
	for (int i = 0; i < datalength; i += 2) {
		SpectromterPix = (NewReadFrambuffer[i] << 8) | (NewReadFrambuffer[i + 1] & 0xff);

		printf("%d", SpectromterPix);
		printf(",");
		Textfile << SpectromterPix;
		Textfile << ",";
	}


	/*
	*********************************************************************
	*********************************************************************
	Write integration time to DISB Register
	*********************************************************************
	*********************************************************************
	*/
	/*
	Setting an integration time of 100 ms.
	Integration time is set in increments of 200 ns, so 10 ms => 50.000 => MSB 0x000 - LBS 0xC350
	*/
	// Setting integration time of Register 10 MSB. 
	SetRegister(ftHandleOutCom, 10, 7);

	// Setting integration time of Register 9 LSB. 
	SetRegister(ftHandleOutCom, 9, 41248);

	// Read the integration time of register 10 MSB. 
	uint16_t MSB_100ms = ReadRegister(ftHandleOutCom, 10);
	// Read the integration time of register 9 MSB. 
	uint16_t LSB_100ms = ReadRegister(ftHandleOutCom, 9);

	IntegrationTime = (MSB_100ms << 16) | (LSB_100ms & 0xffff);

	Textfile << "\n----------------------------------------------------------------------------------------------- ";
	Textfile << "\n";
	Textfile << "\nIntegration time in register in counts: ";
	Textfile << IntegrationTime;
	Textfile << "\nIntegration time in register in milliseconds: ";
	Textfile << IntegrationTime * 0.0002;
	Textfile << " [ms]";

	
	//-------------------------------------------------------------------------------------------------------------------------------
	
	/*
	*********************************************************************
	*********************************************************************
	Trigger exposure via write to Sensor CTRL register.
	*********************************************************************
	*********************************************************************
	*/

	// Writing to the Sensor Control register 8.
	// Perform soft reset of the image transfer buffer
	SetRegister(ftHandleOutCom, 8, 0x10);
	// Trigger an exposure
	SetRegister(ftHandleOutCom, 8, 0x01);

	/*
	*********************************************************************
	*********************************************************************
	Read spectrum from framebuffer.
	*********************************************************************
	*********************************************************************
	*/

	//-------------------------------------------------------------------------------------------------------------------------------
	// Change the chip select from 0 to 1 (SS0 --> SS1) and Framesize from 8-bit bytes to 16-bit
	Textfile << "\n----------------------------------------------------------------------------------------------- ";
	Textfile << "\n";
	printf("\n-----------------------------------------------------------------------------------------------");
	printf("\nSpectrum with 100 ms integration time");
	printf("\n-----------------------------------------------------------------------------------------------");
	printf("\n");

	//Wait for the framebuffer to be full (NumberOfPixReady = LastPixel) or for timeout after 1 second.
	StartTime = clock();
	do
	{
		NumberOfPixReady = ReadRegister(ftHandleOutCom, 12);
		CurrentTime = clock();
		TimeElapsedinSec = (CurrentTime - StartTime) / CLOCKS_PER_SEC;
	} while (LastPixel - NumberOfPixReady != 0 && TimeElapsedinSec < 1);
	if (TimeElapsedinSec >= 1)
	{
		printf("\nTimeout error when capturing spectrum\n");
	}

	//Read the pixel values
	FT4222_SPIMaster_SingleReadWrite(ftHandleOutStream, NewReadFrambuffer.data(), NewWriteFrambuffer.data(), datalength, &SizeTransferred, 1);


	//Stich and print the pixel values
	for (int i = 0; i < datalength; i += 2) {
		SpectromterPix = (NewReadFrambuffer[i] << 8) | (NewReadFrambuffer[i + 1] & 0xff);

		printf("%d", SpectromterPix);
		printf(",");
		Textfile << SpectromterPix;
		Textfile << ",";
	}

	/*
	********************************************************************
	*********************************************************************
	Uninitialize Device Master SPI.
	*********************************************************************
	*********************************************************************
	*/
	FT4222_UnInitialize(ftHandleOutStream);
	/*
	*********************************************************************
	*********************************************************************
	Closing Device FT4222.
	*********************************************************************
	*********************************************************************
	*/
	FT_Close(ftHandleOutStream);

	return 0;
}